@@ -6,8 +6,9 @@ import pathlib
 from typing import Optional
 
 import torch
-from diffusers.pipelines.stable_diffusion.convert_from_ckpt import load_pipeline_from_original_stable_diffusion_ckpt
-from diffusers import StableDiffusionPipeline, DPMSolverMultistepScheduler
+from compel import Compel
+from diffusers import StableDiffusionPipeline, LMSDiscreteScheduler, DPMSolverMultistepScheduler, \
+    EulerAncestralDiscreteScheduler, PNDMScheduler, DDPMScheduler, KDPM2AncestralDiscreteScheduler, DDIMScheduler
 from PIL import Image, ImageDraw, ImageFont
 from diffusers.pipelines.stable_diffusion import StableDiffusionPipelineOutput
 from diffusers.utils import is_xformers_available
@@ -156,6 +157,9 @@ def load_model(repo_id_or_path, prefer_fp16: bool = True, local_files_only: bool
         yaml_file = (os.path.splitext(repo_id_or_path)[0] + ".yaml")
         if not os.path.exists(yaml_file):
             yaml_file = None
+        # this won't work on diffusers 0.15+, need to find a way to reliably support old + new
+        from diffusers.pipelines.stable_diffusion.convert_from_ckpt import \
+                load_pipeline_from_original_stable_diffusion_ckpt
         return load_pipeline_from_original_stable_diffusion_ckpt(repo_id_or_path, original_config_file=yaml_file)
     elif os.path.isdir(repo_id_or_path):
         return StableDiffusionPipeline.from_pretrained(repo_id_or_path)
@@ -169,36 +173,66 @@ def load_model(repo_id_or_path, prefer_fp16: bool = True, local_files_only: bool
         return StableDiffusionPipeline.from_pretrained(repo_id_or_path, revision=revision,
                                                        local_files_only=local_files_only)
 
+def make_scheduler(which: str, config):
+
+    if which == 'ddim':
+        return DDIMScheduler.from_config(config)
+    elif which == "lms":
+        return LMSDiscreteScheduler.from_config(config)
+    elif which == "dpm++":
+        return DPMSolverMultistepScheduler.from_config(config, algorithm_type="dpmsolver++", use_karras_sigmas=False)
+    elif which == "k_dpm++":
+        return DPMSolverMultistepScheduler.from_config(config, algorithm_type="dpmsolver++", use_karras_sigmas=True)
+    elif which == "dpm++_sde":
+        return DPMSolverMultistepScheduler.from_config(config, algorithm_type="sde-dpmsolver++",
+                                                       use_karras_sigmas=False)
+    elif which == "k_dpm++_sde":
+        return DPMSolverMultistepScheduler.from_config(config, algorithm_type="sde-dpmsolver++",
+                                                       use_karras_sigmas=True)
+    elif which == "euler_a":
+        return EulerAncestralDiscreteScheduler.from_config(config, use_karras_sigmas=False)
+    elif which == "k_euler_a":
+        return EulerAncestralDiscreteScheduler.from_config(config, use_karras_sigmas=True)
+    elif which == 'pndm':
+        return PNDMScheduler.from_config(config)
+    elif which == 'ddpm':
+        return DDPMScheduler.from_config(config)
+    elif which == 'k_dpm2_a':
+        return KDPM2AncestralDiscreteScheduler.from_config(config)
+    else:
+        raise ValueError(f"Unrecognized scheduler {which}")
+
 
 def render_row(prompts: list[str],
                negative_prompts: Optional[list[str]],
                seeds: list[int],
                pipeline: StableDiffusionPipeline,
+               scheduler_name: str="dpm++",
                device: str = None,
                batch_size=1,
                sample_w=512,
                sample_h=512,
                cfg=7.5,
-               num_inference_steps=15,  # ddpm++ solver: 15 is typically enough
+               num_inference_steps=15, # for dpm++ solver, 15 is typically enough
                disable_nsfw_checker: bool = False,
+               use_penultimate_clip_layer: bool = False
                ) -> list[Image]:
-    # ddpm++
-    pipeline.scheduler = DPMSolverMultistepScheduler.from_config(pipeline.scheduler.config,
-                                                                 algorithm_type="dpmsolver++")
+    pipeline.scheduler = make_scheduler(which=scheduler_name, config=pipeline.scheduler.config)
     if device is None:
         device = 'cuda' if torch.cuda.is_available() else 'mps' if torch.backends.mps.is_available() else 'cpu'
-    if device == 'cuda':
-        # noinspection PyTypeChecker
-        pipeline = pipeline.to(torch.float16)
-        if is_xformers_available():
-            pipeline.enable_xformers_memory_efficient_attention()
-    pipeline = pipeline.to(device)
+    # noinspection PyTypeChecker
+    pipeline = pipeline.to(device, torch.float16)
+    if is_xformers_available():
+        pipeline.enable_xformers_memory_efficient_attention()
     if disable_nsfw_checker:
         pipeline.safety_checker = disable_nsfw_safety_checker
     images = []
     negative_prompts = negative_prompts or [""] * len(prompts)
     batches = chunk_list(list(zip(prompts, negative_prompts, seeds)), batch_size)
     progress_bar = tqdm(list(batches))
+    compel = Compel(tokenizer=pipeline.tokenizer,
+                    text_encoder=pipeline.text_encoder,
+                    use_penultimate_clip_layer=use_penultimate_clip_layer)
     for batch in progress_bar:
         batch_prompts, batch_negative_prompts, batch_seeds = zip(*batch)
         print(f" - {batch_prompts}")
@@ -295,6 +329,7 @@ def explode_merge_config(merge_config: dict) -> list[dict]:
 
 def render_all(prompts: list[str], negative_prompts: Optional[list[str]], seeds: list[int], cfg: float,
                repo_ids_or_paths: list[str],
+               scheduler: str,
                device: str,
                size: tuple[int, int],
                batch_size: int,
@@ -303,6 +338,7 @@ def render_all(prompts: list[str], negative_prompts: Optional[list[str]], seeds:
                local_files_only: bool = False,
                merge_config: Optional[dict] = None,
                disable_nsfw_checker: bool = False,
+               use_penultimate_clip_layer: Optional[bool] = False, # autodetect if None
                ) -> Image:
     all_images = []
     print(f"{len(prompts)} prompts")
@@ -349,12 +385,14 @@ def render_all(prompts: list[str], negative_prompts: Optional[list[str]], seeds:
                                     negative_prompts=negative_prompts,
                                     seeds=seeds,
                                     pipeline=pipeline,
+                                    scheduler_name=scheduler,
                                     device=device,
                                     batch_size=batch_size,
                                     cfg=cfg,
                                     num_inference_steps=inference_steps,
                                     sample_w=size[0], sample_h=size[1],
-                                    disable_nsfw_checker=disable_nsfw_checker,)
+                                    disable_nsfw_checker=disable_nsfw_checker,
+                                    use_penultimate_clip_layer=use_penultimate_clip_layer)
             all_images += row_images
             save_partial_if_requested()
 
@@ -362,7 +400,7 @@ def render_all(prompts: list[str], negative_prompts: Optional[list[str]], seeds:
             if save_merge_path_prefix is not None:
                 save_half = merge_config.get('save_merge_half', True)
                 if save_half:
-                    pipeline.to(torch.float16)
+                    pipeline.to(pipeline.device, torch.float16)
                 save_merge_path = f"{save_merge_path_prefix}_{merge_index:02d}"
                 pipeline.save_pretrained(save_merge_path)
 
@@ -378,12 +416,14 @@ def render_all(prompts: list[str], negative_prompts: Optional[list[str]], seeds:
                                     negative_prompts=negative_prompts,
                                     seeds=seeds,
                                     pipeline=pipeline,
+                                    scheduler_name=scheduler,
                                     device=device,
                                     batch_size=batch_size,
                                     cfg=cfg,
                                     num_inference_steps=inference_steps,
                                     sample_w=size[0], sample_h=size[1],
-                                    disable_nsfw_checker=disable_nsfw_checker,)
+                                    disable_nsfw_checker=disable_nsfw_checker,
+                                    use_penultimate_clip_layer=use_penultimate_clip_layer)
             all_images += row_images
             save_partial_if_requested()
             del pipeline
@@ -453,6 +493,12 @@ def main():
                         type=int,
                         nargs="+",
                         help="(Optional) Seeds. Specify either one seed to share for all `--prompts`, or as many seeds as there are `--prompts`.")
+    parser.add_argument("--scheduler",
+                        choices=['ddim', "lms", "dpm++", "k_dpm++", "dpm++_sde", "k_dpm++_sde",
+                                 "euler_a", "k_euler_a", 'pndm', 'ddpm', 'k_dpm2_a'],
+                        default="dpm++",
+                        type=str,
+                        help="(Optional, default=dpm++) Scheduler to use.")
     parser.add_argument("--cfg",
                         required=False,
                         type=float,
@@ -504,6 +550,9 @@ def main():
     parser.add_argument("--save_merge_float32",
                         action="store_true",
                         help="(Optional) If saving merges, save with float32 precision (default is float16).")
+    parser.add_argument("--use_penultimate_clip_layer",
+                        action="store_true",
+                        help="(Optional) Use the outputs from penultimate (second to last) CLIP hidden layer. On detected SD2.x models this defaults on, otherwise it defaults off.")
     args = parser.parse_args()
 
 
@@ -548,6 +597,7 @@ def main():
 
     render_all(prompts=prompts,
                negative_prompts=negative_prompts,
+               scheduler=args.scheduler,
                seeds=seeds,
                repo_ids_or_paths=args.repo_ids_or_paths,
                merge_config=merge_config,
@@ -558,7 +608,8 @@ def main():
                inference_steps=args.steps,
                local_files_only=args.local_files_only,
                save_partial_filename=args.output_path,
-               disable_nsfw_checker=args.disable_nsfw_checker,)
+               disable_nsfw_checker=args.disable_nsfw_checker,
+               use_penultimate_clip_layer=args.use_penultimate_clip_layer)
     print(f"grate saved to {args.output_path}")
 
 if __name__ == '__main__':
